/*
 * Model Relinker
 *
 * The idea of this module is to preprocess the imported wc3 model and convert it to
 * an intermediate format for easing the scene rebuild process.
 *
 * This library's purpose is to change model ids for referenced data structures, thus rebuilding the scene is easier.
 * It also formats animations and values to the final product.
 */
 
local OriginalSequences
local OldSequences
local NewSequences

fn TransformTextures &Keys &Textures =
(
	if (Keys.type == #Static) then \
		Keys.data = Textures[Keys.data]
	else
	(
		local data = Keys.data
		local IsSmooth =  isSmoothAnim data.type
		for k in data.keys do
		(
			k.value = Textures[k.value]
			if IsSmooth then
			(
				k.intan = Textures[k.intan]
				k.outtan = Textures[k.outtan]
			)
		)
	)
	Keys
)

fn relinkMaterial m model =
(
	for l in m.layers do
	(
		TransformTextures l.textures model.textures
		if l.tvertexid != undefined then l.tvertexid = model.tvertexAnims[l.tvertexid]
	)
	if m.tvertexid != undefined then m.tvertexid = model.tvertexAnims[m.tvertexid]
)

fn relinkGeosetAnimations g model =
(
	g.geoset = model.geosets[g.geoset]
)

fn relinkGeoset g model =
(
	g.material = model.materials[g.material]
)

fn relinkObject o mdlObjects =
	if o.parent != undefined then o.parent = mdlObjects[o.parent]

-- Animations

struct BasicKey (
	value,
	time,
	lock = false,
	intan,
	outtan
)

fn fixQuaternions keys =
(
	for i = 2 to keys.count do
	(
		local q1 = keys[i].value, q2 = keys[i-1].value
		if (DotQuat q1 q2) < 0 then keys[i].value = -q1
	)
)

fn convertSequences sequences =
(
	newSequences = #()
	oldSequences = #()
	OriginalSequences = sequences
	local start = 10
	for s in Sequences do
	(
		local end = start + s.endTime - s.startTime
		append oldSequences (interval s.startTime s.endTime)
		s.startTime = start
		append newSequences (interval start end)
		start = ((end/10)*10 + 20) as time
		s.endTime = end
	)
)

fn preProcessKey key t lock:false =
(
	if (classOf key == KeyData) then
		BasicKey value:key.value time:t lock:lock intan:undefined outtan:undefined
	else
		BasicKey value:key.value time:t lock:lock intan:key.intan outtan:key.outtan
)

fn createStubKey defaultVal t =
(
	local d = copy defaultVal
	BasicKey value:d time:t lock:false intan:d outtan:d
)

fn optimizeVisibilityKeys keys =
(
	local newKeys = #(keys[1])
	local val = keys[1].value
	for i = 2 to keys.count where keys[i].value != val do
	(
		val = keys[i].value
		append newKeys keys[i]
	)
	newKeys
)

fn preProcessVisibility keys =
(
	local newKeys = #()
	local i = 0
	local sepKeys = getKeysOnSequences keys oldSequences inclusive:true
	for i = 1 to sepKeys.count do
	(
		local isLoop = not OriginalSequences[i].loop
		local keys = sepKeys[i]
		local oldAnim = oldSequences[i]
		local newAnim = newSequences[i]
		local factor1 = newAnim.start - oldAnim.start
		--local durationFactor = (newAnim.end - newAnim.start)/(oldAnim.end - oldAnim.start)
		local start = newAnim.start, end = newAnim.end
		if keys.count == 0 or start != (keys[1].time + factor1) then
				append newKeys (createStubKey 1.0 start)
		for i = 1 to keys.count do
		(
			local k = keys[i]
			local t = k.time + factor1
			append newKeys (preProcessKey k t lock:false)
		)
	)
	newKeys = (optimizeVisibilityKeys newKeys)
	--if (newKeys[1].value < 0.5) then deleteItem newKeys 1
	newKeys
)

fn preProcessKeys keys defaultVal =
(
	local newKeys = #(createStubKey defaultVal 0f)
	local sepKeys = getKeysOnSequences keys oldSequences inclusive:true
	for i = 1 to sepKeys.count do
	(
		local isLoop = not OriginalSequences[i].loop
		local keys = sepKeys[i]
		local oldAnim = oldSequences[i]
		local newAnim = newSequences[i]
		local factor1 = newAnim.start - oldAnim.start
		--local durationFactor = (newAnim.end - newAnim.start)/(oldAnim.end - oldAnim.start)
		local start = newAnim.start, end = newAnim.end
		if keys.count == 0 then
		(
			append newKeys (createStubKey defaultVal start)
			append newKeys (createStubKey defaultVal end)
		)
		else (
			if start != (keys[1].time + factor1) then
				append newKeys (createStubKey defaultVal start)
			for i = 1 to keys.count do
			(
				local k = keys[i]
				local t = k.time + factor1
				local lock = (t == start or t == end)
				append newKeys (preProcessKey k t lock:lock)
			)
			local firstKey = newKeys[newKeys.count - keys.count + 1]
			local lastKey = newKeys[newKeys.count]
			if end != (keys[keys.count].time + factor1) then
				if isLoop then
					append newKeys (preProcessKey firstKey end lock:true)
				else
					append newKeys (preProcessKey lastKey end lock:true)
			else
				lastKey.lock = true
		)
	)
	newKeys
)

fn preProcessEventKeys keys =
(
	local newKeys = #()
	local sepKeys = getKeysOnSequences keys oldSequences inclusive:true
	for i = 1 to sepKeys.count do
	(
		local keys = sepKeys[i]
		local oldAnim = oldSequences[i]
		local newAnim = newSequences[i]
		local factor1 = newAnim.start - oldAnim.start
		--local durationFactor = (newAnim.end - newAnim.start)/(oldAnim.end - oldAnim.start)
		local start = newAnim.start, end = newAnim.end
		if keys.count != 0 then
		(
			for i = 1 to keys.count do
			(
				local k = keys[i]
				local t = k.time + factor1
				append newKeys (preProcessKey k t lock:false)
			)
		)
	)
	newKeys
)

fn preProcessAnim animData defaultVal =
(
	if animData.globalSequenceId == undefined then
		preProcessKeys animData.keys defaultVal
	else animData.keys
)

fn convertAnimations o =
(
	--fix dot quat
	local trans = o.transformations
	if trans.rotation != undefined then
	(
		trans.rotation.keys = preProcessAnim trans.rotation (quat 0 0 0 1)
		fixQuaternions trans.rotation.keys
	)
	if trans.translation != undefined then
		trans.translation.keys = preProcessAnim trans.translation [0,0,0]
	if trans.scale != undefined then
		trans.scale.keys = preProcessAnim trans.scale [1,1,1]
)

fn preProcessParamAnim param default =
(
	if (param.type == #Animated) then
		param.data.keys = (preProcessAnim param.data default)
)

fn convertLightAnimations l =
(
	preProcessParamAnim l.color [255,255,255]
	preProcessParamAnim l.intensity 0.0
	preProcessParamAnim l.ambColor [255,255,255]
	preProcessParamAnim l.ambIntensity 0.0
	preProcessParamAnim l.attenuationStart 0.0
	preProcessParamAnim l.attenuationEnd 0.0
)

fn convertParticleEmitter1Animations p =
(
	preProcessParamAnim p.emissionRate 0.0
	preProcessParamAnim p.gravity 0.0
	preProcessParamAnim p.longitude 0.0
	preProcessParamAnim p.latitude 0.0
	preProcessParamAnim p.lifespan 0.0
)

fn convertParticleEmitter2Animations p =
(
	preProcessParamAnim p.speed 0.0
	preProcessParamAnim p.gravity 0.0
	preProcessParamAnim p.variation 0.0
	preProcessParamAnim p.latitude 0.0
	preProcessParamAnim p.emissionRate 0.0
	preProcessParamAnim p.width 0.0
	preProcessParamAnim p.length 0.0
	if (p.visibility.type == #animated) then
	(
		if p.visibility.data.type == #NonInterp then
			p.visibility.data.keys = (preProcessVisibility p.visibility.data.keys)
		else
			preProcessParamAnim p.visibility 0.0
	)
)

fn convertRibbonEmitterAnimations r =
(
	preProcessParamAnim r.heightAbove 0.0
	preProcessParamAnim r.heightBelow 0.0
	preProcessParamAnim r.alpha 0.0
	preProcessParamAnim r.textureSlot 0.0
	preProcessParamAnim r.color [0,0,0]
)

fn converEventObjectAnimations o =
(
	if o.eventTrack.globalSequenceId == undefined then
		o.eventTrack.keys = preProcessEventKeys o.eventTrack.keys
)

fn convertParameterAnimations o =
(
	local s = o.sub
	if (isProperty s #visibility) and (s.visibility.type == #Animated) then
		s.visibility.data.keys = (preProcessVisibility s.visibility.data.keys)
)

fn convertGeosetAnimations geoAnim =
(
	if (geoAnim.alpha.type == #animated) then
		geoAnim.alpha.data.keys = (preProcessVisibility geoAnim.alpha.data.keys)
	if (geoAnim.usesColor and geoAnim.color.type == #animated) then
		geoAnim.color.data.keys = (preProcessAnim geoAnim.color.data white)
)

fn RelinkParticleEmitter2 p textures =
(
	p.sub.texture = textures[p.sub.texture]
)

fn splitTvertexAnimation tv =
(
	if (tv.transformation.translation != undefined) then
	(
		local u = AnimationData ()
		local v = AnimationData ()
		local w = AnimationData ()
		local t = tv.transformation.translation.type
		local g = tv.transformation.translation.globalSequenceId
		local result = #(u,v,w)
		local keys = tv.transformation.translation.keys
		local isSmooth = (t == #Bezier or t == #Hermite)
		local s = keys.count
		for i = 1 to 3 do
		(
			local r = result[i]
			r.type = t
			r.globalSequenceId = g
			local newKeys = #()
			newKeys[s] = undefined
			--local prev = 1.0
			for j = 1 to s do
			(
				local k = keys[j]
				local p = k.value[i]
				if isSmooth then
					newKeys[j] = SmoothKeyData k.time p (k.intan[i]) (k.outtan[i])
				else newKeys[j] = KeyData k.time p
				--prev = p
			)
			r.keys = newKeys
			preProcessAnim r 0.0
		)
		tv.transformation.translation = result
	)
)

fn relinkModel &model =
(
	progressDownLevel 5 1
	progressShowItem "Linking Materials"
	for m in model.materials do
		relinkMaterial m model
	progressUpLevel ()
	progressDownLevel 10 1
	progressShowItem "Linking Geosets"
	for g in model.geosets do
		relinkGeoset g model
	progressUpLevel ()
	progressDownLevel 5 1
	progressShowItem "Linking Geoset Animations"
	for g in model.geosetAnims do
		relinkGeosetAnimations g model
	progressUpLevel ()
	progressDownLevel 10 1
	progressShowItem "Linking Objects"
	local mdlObjects = #()
	for n in model.bones do
		mdlObjects[n.objId] = n
	for n in model.helpers do
		mdlObjects[n.objId] = n
	for n in model.lights do
		mdlObjects[n.objId] = n
	for n in model.attachments do
		mdlObjects[n.objId] = n
	for n in model.events do
		mdlObjects[n.objId] = n
	for n in model.particleEmitters1 do
		mdlObjects[n.objId] = n
	for n in model.particleEmitters2 do
		mdlObjects[n.objId] = n
	for n in model.ribbonEmitters do
		mdlObjects[n.objId] = n
	for n in model.CollisionShapes do
		mdlObjects[n.objId] = n
	for o in mdlObjects where o != undefined do
		relinkObject o mdlObjects
	model.ex = mdlObjects
	-- Particular Relinks
	for p in model.particleEmitters2 do
		RelinkParticleEmitter2 p model.textures
	for r in model.ribbonEmitters do
	(
		relinkGeoset r.sub model
		convertRibbonEmitterAnimations r.sub
	)
	progressUpLevel ()
	progressDownLevel 20 1
	progressShowItem "Converting Animations"
	-- convert Animations
	for t in model.tvertexAnims do
		splitTvertexAnimation t
	convertSequences model.sequences
	gc ()
	for o in mdlObjects where o != undefined do
	(
		convertAnimations o
		convertParameterAnimations o
	)
	for g in model.geosetAnims do
		convertGeosetAnimations g
	for l in model.lights do
		convertLightAnimations l.sub
	for p in model.particleEmitters1 do
		convertParticleEmitter1Animations p.sub
	for p in model.particleEmitters2 do
		convertParticleEmitter2Animations p.sub
	for r in model.ribbonEmitters do
		convertRibbonEmitterAnimations r.sub
	for o in model.events do
		converEventObjectAnimations o.sub
	progressUpLevel ()
)